#include "ntddk.h"

typedef unsigned int DWORD;

typedef struct _SYSTEM_MODULE_INFORMATION {
    ULONG    Reserved[2];
    PVOID    Base;
    ULONG    Size;
    ULONG    Flags;
    USHORT   Index;
    USHORT   Unknown;
    USHORT   LoadCount;
    USHORT   ModuleNameOffset;
    CHAR     ImageName[256];
} SYSTEM_MODULE_INFORMATION, *PSYSTEM_MODULE_INFORMATION;

typedef struct {
    DWORD    dwNumberOfModules;
    SYSTEM_MODULE_INFORMATION    smi;
} MODULES, *PMODULES;

#define PREAMBLE_SIZE 10
#pragma pack(1)
typedef struct SDEntry {
	unsigned int *Base;
	unsigned int *dummy;
	unsigned int num;
	unsigned char *ptable;
} SSDT_Entry;

typedef struct fixup_entry {
	DWORD func_addr;
	unsigned char func_preamble[PREAMBLE_SIZE];
} fu_entry;
#pragma pack()

__declspec(dllimport) SSDT_Entry KeServiceDescriptorTable;
__declspec(dllimport) NTSTATUS __stdcall ZwQuerySystemInformation(DWORD, DWORD, DWORD, DWORD);
#define NT_SUCCESS(Status) ((NTSTATUS)(Status) >= 0)
#define MAGIC 0x30090000

// Cerca il modulo dove vogliamo far passare i parametri
DWORD FindExe(unsigned char *string)
{
	if (_stricmp(string, "ntdll.dll") == 0)
		return 1;
	return 0;
}

DWORD fix_preamble(unsigned char *func_addr, unsigned char *func_preamble)
{
	PMDL func_mdl;
	unsigned char *safe_func_addr;
	DWORD i;
	
	if (func_preamble[0] == 0)
		return 1;
		
	if ( !(func_mdl = IoAllocateMdl(func_addr, PREAMBLE_SIZE, FALSE, TRUE, NULL)) )
		return  0;
	try {
		MmProbeAndLockPages(func_mdl, KernelMode, IoWriteAccess);
	} except(EXCEPTION_EXECUTE_HANDLER) {
		IoFreeMdl(func_mdl);
		return  0;
	}
	if ( !(safe_func_addr = MmGetSystemAddressForMdlSafe(func_mdl, HighPagePriority)) ) {
		MmUnlockPages(func_mdl);
		IoFreeMdl(func_mdl);
		return  0;
	}
	
	for (i=0; i<PREAMBLE_SIZE; i++) 
		safe_func_addr[i] = func_preamble[i];
	
	MmUnlockPages(func_mdl);
	IoFreeMdl(func_mdl);
	return  1;
}



NTSTATUS DriverEntry( IN PDRIVER_OBJECT dobj, IN PUNICODE_STRING regpath)
{
	NTSTATUS rc;
	PMDL uspace_mdl;
	DWORD *kusp;
	DWORD *str_addr;
	DWORD pModules;
	PMODULES pModStruct;
	PSYSTEM_MODULE_INFORMATION psmi;
	DWORD image_base, kernel_base;
	fu_entry *old_sdt;
	DWORD entry_num;
	DWORD i, len;
	DWORD *mapped_sys_table;	

	// Prende la lista dei moduli
	pModules = (DWORD)&pModules;
	rc = ZwQuerySystemInformation(11, pModules, 4, (DWORD)&len);
	if (rc == STATUS_INFO_LENGTH_MISMATCH) {
		pModules = (DWORD)ExAllocatePool(NonPagedPool, len);
    		if (!pModules)
    			return STATUS_UNSUCCESSFUL;
		rc = ZwQuerySystemInformation(11, pModules, len, (DWORD)NULL);
		if (!NT_SUCCESS(rc)) {
			ExFreePool((PVOID)pModules);
			return STATUS_UNSUCCESSFUL;
		}
    } else
		return STATUS_UNSUCCESSFUL;

	// Prende la base del kernel...
	pModStruct = (PMODULES)pModules;
	kernel_base = (DWORD)pModStruct->smi.Base;
	psmi = &(pModStruct->smi);
	psmi++; 
	image_base = 0;
	// ...e di ntdll.dll
	for (i=1; i<pModStruct->dwNumberOfModules; i++, psmi++) 
		if (FindExe(psmi->ImageName + psmi->ModuleNameOffset))
			image_base = (DWORD)psmi->Base;	

	ExFreePool((PVOID)pModules);
	if (!image_base)
		return  STATUS_UNSUCCESSFUL;

	// Mappa l'header di ntdll	
	str_addr = (unsigned int *)(image_base + (7*4));
	if ( !(uspace_mdl = IoAllocateMdl(str_addr, 8,  FALSE, TRUE, NULL)) )
		return  STATUS_UNSUCCESSFUL;

	try {
		MmProbeAndLockPages(uspace_mdl, UserMode, IoReadAccess);
	} except(EXCEPTION_EXECUTE_HANDLER) {
		IoFreeMdl(uspace_mdl);
		return  STATUS_UNSUCCESSFUL;
	}
	if ( !(kusp = MmGetSystemAddressForMdlSafe(uspace_mdl, HighPagePriority)) ) {
		MmUnlockPages(uspace_mdl);
		IoFreeMdl(uspace_mdl);
		return  STATUS_UNSUCCESSFUL;
	}
	// Legge dall'header l'indirizzo della vecchia sdt
	// e il numero di entry
	old_sdt = (fu_entry *)(kusp[0]);
	entry_num = kusp[1];
	MmUnlockPages(uspace_mdl);
	IoFreeMdl(uspace_mdl);
	// Check di consistenza
	if ( (entry_num & 0xFFFF) > KeServiceDescriptorTable.num && (entry_num & 0xFFFF0000) != MAGIC) 
		return  STATUS_UNSUCCESSFUL;
	entry_num &= 0xFFFF;
	
	// Mappa la vecchia SDT in user space
	if ( !(uspace_mdl = IoAllocateMdl(old_sdt, entry_num*sizeof(fu_entry),  FALSE, TRUE, NULL)) )
		return  STATUS_UNSUCCESSFUL;
	try {
		MmProbeAndLockPages(uspace_mdl, UserMode, IoReadAccess);
	} except(EXCEPTION_EXECUTE_HANDLER) {
		IoFreeMdl(uspace_mdl);
		return  STATUS_UNSUCCESSFUL;
	}
	if ( !(old_sdt = MmGetSystemAddressForMdlSafe(uspace_mdl, HighPagePriority)) ) {
		MmUnlockPages(uspace_mdl);
		IoFreeMdl(uspace_mdl);
		return  STATUS_UNSUCCESSFUL;
	}

	// Sovrascrive le entry
	mapped_sys_table = KeServiceDescriptorTable.Base;
	for (i=0; i<entry_num; i++) {
		mapped_sys_table[i] = old_sdt[i].func_addr + kernel_base;
		// Aggiusta il preambolo
		fix_preamble((unsigned char *)mapped_sys_table[i], old_sdt[i].func_preamble);
	}

	MmUnlockPages(uspace_mdl);
	IoFreeMdl(uspace_mdl);

	return STATUS_SUCCESS;
}